/**
 **	Actors.h - Game actors.
 **
 **	Written: 11/3/98 - JSF
 **/

/*
Copyright (C) 1998  Jeffrey S. Freedman

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef INCL_ACTORS
#define INCL_ACTORS	1

#include "objs.h"

/*
 *	An actor:
 */
class Actor : public Moveable_sprite
	{
	Actor *next, *prev;		// Next, prev. in vicinity.
public:
	Actor(int shapenum) : Moveable_sprite(shapenum)
		{
		next = prev = this;
		}
	void add_after_this(Actor *a)	// Add another actor after this.
		{
					// Remove from current chain.
		a->next->prev = a->prev;
		a->prev->next = a->next;
		a->next = next;		// Add to this one.
		a->prev = this;
		next->prev = a;
		next = a;
		}
	Actor *get_next()
		{ return next; }			
	};

/*
 *	The standard human sprites in the game use the same frame #'s for
 *	animation.
 */
class Human_actor : public Actor
	{
public:
	Human_actor(int shapenum);
	};

/*
 *	Here's an actor that's just hanging around an area.
 */
class Area_actor : public Human_actor
	{
	timeval next_change;		// When to change motion.
public:
	Area_actor(int shapenum) : Human_actor(shapenum)
		{
		next_change.tv_sec = next_change.tv_usec = 0;
		}
					// Figure next frame location.
	virtual int next_frame(timeval& time,
		int& new_cx, int& new_cy, int& new_sx, int& new_sy,
		int& new_frame);
	};

#endif
