/**
 **	Exult.cc - X-windows Ultima7 map browser.
 **
 **	Written: 7/22/98 - JSF
 **/

/*
Copyright (C) 1998  Jeffrey S. Freedman

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <string.h>
#include <fstream.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "imagewin.h"
#include "actors.h"
#include "utils.h"
#include <X11/Xutil.h>
#define XK_MISCELLANY
#define XK_LATIN1
#include <X11/keysymdef.h>
#include "exult.h"
#include "items.h"

/*
 *	A handy breakpoint.
 */

static void Breakpoint
	(
	)
	{
	return;
	}

/*
 *	Convert a buffer to upper-case.
 *
 *	Output:	->original buffer, changed to upper case.
 */

static char *To_upper
	(
	char *buf
	)
	{
	char *ret = buf;
	while (*buf)
		{
		*buf = toupper(*buf);
		buf++;
		}
	return (ret);
	}

/*
 *	Create game window.
 */

Game_window::Game_window
	(
	Display *dpy
	) : display(dpy), chunkx(0), chunky(0), painted(0), focus(1),
	    brightness(100), 
	    skip_lift(16), debug(0), shapewin(0), quitting_time(0),
	    main_actor(247),		// Someone to move around.
	    main_actor_inside(0), mode(intro), showing_item(0)
	{
	u7open(chunks, "static/u7chunks");
	u7open(shapes, "static/shapes.vga");
	u7open(u7map, "static/u7map");
	Setup_item_names();		// Set up list of item names.
					// Get screen info.
	screen_num = DefaultScreen(display);
	unsigned int display_width = DisplayWidth(display, screen_num);
	unsigned int display_height = DisplayHeight(display, screen_num);
					// Create window 1/2 of screen size.
	win = Image_window::create(display, screen_num,
		display_width/2, display_height/2,
		BlackPixel(display, screen_num), 
		WhitePixel(display, screen_num));
					// Get 12-point font.
	font12 = win->open_font("avatar.ttf", 12);
					// Set title.
	win->set_title("Exult Ultima7 Viewer");
					// Indicate events we want.
	win->select_input(ExposureMask | KeyPressMask |
		PointerMotionHintMask |
		ButtonPressMask | ButtonReleaseMask | Button3MotionMask | 
		FocusChangeMask | StructureNotifyMask);

	struct timeval timer;
	gettimeofday(&timer, 0);	// Get time of day.
	srand(timer.tv_usec);		// Use it to seed rand. generator.
	shapes.seekg(0x54);		// Get # of shapes.
	num_shapes = Read4(shapes);
					// Set up lists of pointers.
	shapes_read = new Shape **[num_shapes];
	memset((char *) shapes_read, 0, num_shapes * sizeof(Shape **));
					// Clear object lists, flags.
	memset((char *) objects, 0, sizeof(objects));
	memset((char *) schunk_read, 0, sizeof(schunk_read));
	get_palette(0);			// Try first palette.
	brighten(20);			// Brighten 20%.
	win->map();			// Now display window.
	}

/*
 *	Deleting game window.
 */

Game_window::~Game_window
	(
	)
	{
	win->close_font(font12);
	delete win;
	delete shapewin;
	}

/*
 *	Create a window for showing individual shapes.
 */

void Game_window::open_shape_window
	(
	)
	{
	if (shapewin)
		return;			// Already done.
					// Make it 128x128.
	shapewin = Image_window::create(display, screen_num,
		128, 128,
		BlackPixel(display, screen_num), 
		WhitePixel(display, screen_num));
					// Indicate events we want.
	shapewin->select_input(ExposureMask |
		ButtonPressMask | StructureNotifyMask);
					// Set title.
	shapewin->set_title("Exult Shapes");
	shapewin->map();		// Show it on screen.
	get_palette(0, brightness);	// Re-read palette.
	}

/*
 *	Abort.
 */

void Game_window::abort
	(
	char *msg,
	...
	)
	{
	va_list ap;
	va_start(ap, msg);
	char buf[512];
	vsprintf(buf, msg, ap);		// Format the message.
	cerr << "Exult (fatal): " << buf << '\n';
	delete this;
	exit(-1);
	}

/*
 *	Resize event occurred.
 */

void Game_window::resized
	(
	Window xwin,
	unsigned int neww, 
	unsigned int newh
	)
	{ 			
					// Pick right window.
	if (shapewin && xwin == shapewin->get_win())
		{
		shapewin->resized(display, screen_num, neww, newh);
		shapewin_paint(shapewin_cur);
		}
	else
		{
		win->resized(display, screen_num, neww, newh);
		paint();
		}
	}

/*
 *	Open a file, trying the original name (lower case), and the upper
 *	case version of the name.  If it can't be opened either way, we
 *	abort with an error message.
 *
 *	Output:	0 if couldn't open and dont_abort == 1.
 */

int Game_window::u7open
	(
	ifstream& in,			// Input stream to open.
	char *fname,			// May be converted to upper-case.
	int dont_abort			// 1 to just return 0.
	)
	{
	in.open(fname);			// Try to open original name.
	if (!in.good())			// No good?  Try upper-case.
		{
		char upper[512];
		in.open(To_upper(strcpy(upper, fname)));
		if (!in.good())
			abort("Can't open '%s'.", fname);
		}
	}
#if 1
/*
 *	Show the shape the mouse was clicked on.
 */

void Game_window::debug_shape
	(
	Window xwin,			// Which window.
	int x, int y			// Point on screen.
	)
	{
					// Clicked in shapes window?
	if (shapewin && xwin == shapewin->get_win())
		{			// Show next shape in list.
		shapewin_paint(shapewin_cur + 1);
		shapewin->show();
		return;
		}
	cout << "Mouse is at " << x << ',' << y << '\n';
	shapewin_cur = 0;
	for (int lift = 0; lift < 16; lift ++)
		find_debug_shapes(lift, x, y);
					// Set EOL entry.
	shapewin_objs[shapewin_cur] = Game_object();
	shapewin_cur = 0;
	if (!shapewin)			// Open if first time.
		open_shape_window();
	else
		{
		shapewin_paint(0);	// Already open, so update it.
		shapewin->show();
		}
	}

/*
 *	Find objects at a given position on the screen with a given lift.
 *
 *	Output:	Objects are stored in shapewin_objs list.
 */

void Game_window::find_debug_shapes
	(
	int lift,			// Look for objs. with this lift.
	int x, int y			// Pos. on screen.
	)
	{


	x += 4*lift;			// Look for given lift.
	y += 4*lift;
	int cx = x/chunksize;		// Figure chunk #'s within window.
	int cy = y/chunksize;
	int xoff = cx*chunksize;
	int yoff = cy*chunksize;
					// Figure shape within chunk.
	int sx = (x - xoff)/8;
	int sy = (y - yoff)/8;
	cx += chunkx;			// Figure abs. chunk coords.
	cy += chunky;
	Game_object *obj;
	Chunk_object_list& olist = objects[cx][cy];
	for (obj = olist.get_first(); obj; obj = olist.get_next(obj))
		{
		if (obj->get_lift() != lift)
			continue;
		int shapex = obj->get_shape_pos_x();
		int shapey = obj->get_shape_pos_y();
		if (shapex != sx || shapey != sy)
			continue;
		int shapenum = obj->get_shapenum();
		int framenum = obj->get_framenum();
		cout << "\tShape " << shapenum << 
				", frame " << framenum <<
				" at chunk coord (" << cx << ',' <<
				cy <<
				"), shape coord (" << shapex << ',' <<
				shapey << ") with lift " << 
				obj->get_lift() << '\n';
		Shape *shape = get_shape(shapenum, framenum);
		if (shape)
			{
			Rectangle r = get_shape_rect(obj, cx, cy);
			cout << "\t\t(w0,w1,h0,h1) = (" << shape->xleft << 
				',' <<
				shape->xright  << ',' << shape->yabove << 
				',' <<
				shape->ybelow << ")";
			cout << "  Rect = (" << r.x << ',' << r.y << ',' <<
				r.w << ',' << r.h << ")\n";
			}
					// Store shapes in list.
		shapewin_objs[shapewin_cur++] = *obj;
		}
	}

/*
 *	Set the shape in the shape window.
 */

void Game_window::shapewin_paint
	(
	int num				// Number within list to show.
	)
	{
	shapewin_cur = num;
					// Watch for end of list.
	if (num >= sizeof(shapewin_objs)/sizeof(shapewin_objs[0]) ||
	    shapewin_objs[num].is_eol())
		shapewin_cur = 0;
	Game_object *obj = &shapewin_objs[shapewin_cur];
	if (obj->is_eol())
		return;			// Nothing's valid.
	char title[120];		// Set title.
	sprintf(title, "Shape %d, frame %d", obj->get_shapenum(),
						obj->get_framenum());
	shapewin->set_title(title);
	shapewin->fill8(0);		// Fill (with black).
	paint_shape(shapewin, 64, 64, obj->get_shapenum(), obj->get_framenum());
	}
#endif

/*
 *	Get a shape onto the screen.
 */

void Game_window::paint_shape
	(
	Image_window *iwin,		// Window to display it in.
	int xoff, int yoff,		// Where to draw in window.
	int shapenum,			// Shape #.
	int framenum			// Frame # within shape.
	)
	{
	Shape *shape = get_shape(shapenum, framenum);
	if (!shape)
		return;
	if (!shape->rle)		// Not RLE?
		iwin->copy8(shape->data, 8, 8, xoff, yoff);
	else
					// Get compressed data.
		get_rle_shape(iwin, *shape, xoff + 8, yoff + 8);
	}

/*
 *	Show a Run-Length_Encoded shape.
 */

void Game_window::get_rle_shape
	(
	Image_window *iwin,		// Window to draw it in.
	Shape& shape,			// Shape to display.
	int xoff, int yoff		// Where to show in iwin.
	)
	{
	unsigned char *in = shape.data;	// Point to data.
	int scanlen;
	while ((scanlen = Read2(in)) != 0)
		{
					// Get length of scan line.
		int encoded = scanlen&1;// Is it encoded?
		scanlen = scanlen>>1;
		short scanx = Read2(in);
		short scany = Read2(in);
		if (!encoded)		// Raw data?
			{
			iwin->copy8(in, scanlen, 1,
					xoff + scanx, yoff + scany);
			in += scanlen;
			continue;
			}
		for (int b = 0; b < scanlen; )
			{
			unsigned char bcnt = *in++;
					// Repeat next char. if odd.
			int repeat = bcnt&1;
			bcnt = bcnt>>1;	// Get count.
			if (repeat)
				{
				unsigned char pix = *in++;
				iwin->fill8(pix, bcnt, 1,
					xoff + scanx + b, yoff + scany);
				}
			else		// Get that # of bytes.
				{
				iwin->copy8(in, bcnt, 1,
					xoff + scanx + b, yoff + scany);
				in += bcnt;
				}
			b += bcnt;
			}
		}
	}

/*
 *	Get the map objects and scenery for a superchunk.
 */

void Game_window::get_map_objects
	(
	int schunk			// Superchunk # (0-143).
	)
	{
	u7map.seekg(schunk * 16*16*2);	// Get to desired chunk.
	unsigned char buf[16*16*2];
	u7map.read(buf, sizeof(buf));	// Read in the chunk #'s.
	int scy = 16*(schunk/12);	// Get abs. chunk coords.
	int scx = 16*(schunk%12);
	unsigned char *mapdata = buf;
					// Go through chunks.
	for (int cy = 0; cy < 16; cy++)
		for (int cx = 0; cx < 16; cx++)
			{
					// Get chunk #.
			int chunk_num = *mapdata++;
			chunk_num += (*mapdata++) << 8;
			get_chunk_objects(scx + cx, scy + cy, chunk_num);
			}
	}

static int rlecount = 0;	//+++++++++++DEBUGGING.

/*
 *	Read in graphics data into window's image.
 */

void Game_window::get_chunk_objects
	(
	int cx, int cy,			// Chunk index within map.
	int chunk_num			// Desired chunk # within file.
	)
	{
	chunks.seekg(chunk_num * 512);	// Get to desired chunk.
	unsigned char buf[16*16*2];	// Read in 16x16 2-byte shape #'s.
	chunks.read(buf, sizeof(buf));
	unsigned char *data = &buf[0];
	Game_object *rles = 0;		// We'll store RLE shapes list here.
	Game_object *last_rle;		// ->last one we added.
	int rlenum = 0;			// An index into rles.
					// Get list we'll store into.
	Chunk_object_list& olist = objects[cx][cy];
					// A chunk is 16x16 shapes.
	for (int shapey = 0; shapey < 16; shapey++)
		for (int shapex = 0; shapex < 16; shapex++)
			{
			ShapeID id(data[0], data[1]);
			Shape *shape = get_shape(id);
			if (shape->rle)
				{
				Game_object *obj = new Game_object(
					data[0], data[1], shapex, shapey);
				if (rles)
					last_rle->set_next(obj);
				else
					rles = obj;
				last_rle = obj;
				rlecount++;
				}
			else		// Flat.
				olist.set_flat(shapex, shapey, id);
			data += 2;
			}
	if (rles)
		{
		last_rle->set_next(0);	// Finish list.
		olist.set(rles);	// Store list (already sorted by
					//   (y, x)).
		}
	}


/*
 *	Read in the objects for a superchunk from one of the "u7ifix" files.
 */

void Game_window::get_ifix_objects
	(
	int schunk			// Superchunk # (0-143).
	)
	{
	char fname[128];		// Set up name.
	strcpy(fname, "static/u7ifix");
	int len = strlen(fname);
	fname[len] = '0' + schunk/16;
	int lb = schunk%16;
	fname[len + 1] = lb < 10 ? ('0' + lb) : ('a' + (lb - 10));
	fname[len + 2] = 0;
	ifstream ifix;			// There it is.
	u7open(ifix, fname);
	int scy = 16*(schunk/12);	// Get abs. chunk coords.
	int scx = 16*(schunk%12);
					// Go through chunks.
	for (int cy = 0; cy < 16; cy++)
		for (int cx = 0; cx < 16; cx++)
			{
					// Get to index entry for chunk.
			int chunk_num = cy*16 + cx;
			ifix.seekg(0x80 + chunk_num*8);
					// Get location, length.
			long shapesoff = Read4(ifix);
			if (!shapesoff)	// Nothing there?
				continue;
			unsigned long shapeslen = Read4(ifix);
			get_ifix_chunk_objects(ifix, shapesoff, shapeslen/4,
				scx + cx, scy + cy);
			}
	}

/*
 *	Get the objects from one ifix chunk entry onto the screen.
 */

void Game_window::get_ifix_chunk_objects
	(
	ifstream& ifix,
	long filepos,			// Where chunk's data lies.
	int cnt,			// # entries (objects).
	int cx, int cy			// Absolute chunk #'s.
	)
	{
	ifix.seekg(filepos);		// Get to actual shape.
					// Get buffer to hold entries' indices.
	unsigned char *entries = new unsigned char[4*cnt];
	unsigned char *ent = entries;
	ifix.read(entries, 4*cnt);	// Read them in.
	for (int i = 0; i < cnt; i++, ent += 4)
		{
		Game_object *obj = new Game_object(ent);
		objects[cx][cy].add(obj);
		rlecount++;		//++++++++++++++++++++DEBUGGING
		}
	delete[] entries;		// Done with buffer.
	}

/*
 *	Read in the objects for a superchunk from one of the "u7ireg" files.
 *	(These are the moveable objects.)
 */

void Game_window::get_ireg_objects
	(
	int schunk			// Superchunk # (0-143).
	)
	{
	char fname[128];		// Set up name.
	strcpy(fname, "gamedat/u7ireg");
	int len = strlen(fname);
	fname[len] = '0' + schunk/16;
	int lb = schunk%16;
	fname[len + 1] = lb < 10 ? ('0' + lb) : ('a' + (lb - 10));
	fname[len + 2] = 0;
	ifstream ireg;			// There it is.
	if (!u7open(ireg, fname, 1))
		return;			// Just don't show them.
	int scy = 16*(schunk/12);	// Get abs. chunk coords.
	int scx = 16*(schunk%12);
	int entlen;			// Gets entry length.
					// Go through entries.
	while (((entlen = Read1(ireg), ireg.good())))
		{
		if (!entlen || entlen == 1)
			continue;	// Skip 0's & ends of containers.
		if (entlen != 6 && entlen != 12)
			{		// 
			long pos = ireg.tellg();
			ireg.seekg(pos + entlen);
			continue;	// Only know these two types.
			}
		unsigned char entry[12];// Get entry.
		ireg.read(entry, entlen);
		int cx = entry[0] >> 4;	// Get chunk indices within schunk.
		int cy = entry[1] >> 4;
					// Get shape #'s.
		int shapex = entry[0] & 0xf;
		int shapey = entry[1] & 0xf;
					// Not sure about this:  maybe entry[5]
		unsigned int lift = (entlen == 6 ? entry[4] : entry[9]) >> 4;
					// Create obj.
		Game_object *obj = new Game_object(
				entry[2], entry[3], shapex, shapey, lift);
		objects[scx + cx][scy + cy].add(obj);
		rlecount++;		//++++++++++++++++++++DEBUGGING
		}
	}

/*
 *	Read in the objects in a superchunk.
 */

void Game_window::get_superchunk_objects
	(
	int schunk			// Superchunk #.
	)
	{
	get_map_objects(schunk);	// Get map objects/scenery.
	get_ifix_objects(schunk);	// Get objects from ifix.
	get_ireg_objects(schunk);	// Get moveable objects.
	schunk_read[schunk] = 1;	// Done this one now.
// cout << "Rlecount = " << rlecount << '\n';
	}

/*
 *	Put the actor(s) in the world.
 */

void Game_window::init_actors
	(
	)
	{
	main_actor.move(chunkx, chunky + 1, 
				&objects[chunkx][chunky + 1], 14, 8, -1);
#if 1
					//+++++Add another (Iolo).
	Actor *act2 = new Area_actor(154);
	act2->move(chunkx + 1, chunky + 1, 
				&objects[chunkx + 1][chunky + 1], 6, 10, -1);
	main_actor.add_after_this(act2);
					// And another (female Avatar).
	Actor *act3 = new Area_actor(989);
	act3->move(chunkx + 1, chunky + 1, 
				&objects[chunkx + 1][chunky + 1], 7, 14, -1);
	act2->add_after_this(act3);
#endif
	}


/*
 *	Paint a rectangle in the window by pulling in vga chunks.
 */

void Game_window::paint
	(
	int x, int y, int w, int h	// Rectangle to cover.
	)
	{
	if (!win->ready())
		return;
	win->set_clip(x, y, w, h);	// Clip to this area.
					// Which chunks to start with:
					// Watch for shapes 1 chunk to left.
	int start_chunkx = chunkx + x/chunksize - 1;
	if (start_chunkx < 0)
		start_chunkx = 0;
	int start_chunky = chunky + y/chunksize - 1;
	if (start_chunky < 0)
		start_chunky = 0;
	int startx = (start_chunkx - chunkx) * chunksize;
	int starty = (start_chunky - chunky) * chunksize;
	int stopx = x + w;
	int stopy = y + h;
					// Go 1 chunk past end.
	int stop_chunkx = chunkx + stopx/chunksize + 
						(stopx%chunksize != 0) + 1;
	if (stop_chunkx > num_chunks)
		stop_chunkx = num_chunks;
	int stop_chunky = chunky + stopy/chunksize + 
						(stopy%chunksize != 0) + 1;
	if (stop_chunky > num_chunks)
		stop_chunky = num_chunks;
	int xoff, yoff;			// Offsets on screen.
	int cx, cy;			// Chunk #'s.
					// Read in "map", "ifix" objects for
					//  all visible superchunks.
	for (cy = start_chunky; cy < stop_chunky; )
		{
		int schunky = cy/16;	// 16 chunks/superchunk.
					// How far to bottom of superchunk?
		int num_chunks_y = 16 - (cy%16);
		if (cy + num_chunks_y > stop_chunky)
			{
			num_chunks_y = stop_chunky - cy;
			if (num_chunks_y <= 0)
				break;	// Past end of world.
			}
		for (cx = start_chunkx; cx < stop_chunkx; )
			{
			int schunkx = cx/16;
			int num_chunks_x = 16 - (cx%16);
			if (cx + num_chunks_x > stop_chunkx)
				{
				num_chunks_x = stop_chunkx - cx;
				if (num_chunks_x <= 0)
					break;
				}
					// Figure superchunk #.
			int schunk = 12*schunky + schunkx;
					// Read it if necessary.
			if (!schunk_read[schunk])
				get_superchunk_objects(schunk);
					// Increment x coords.
			cx += num_chunks_x;
			}
					// Increment y coords.
		cy += num_chunks_y;
		}
	if (!main_actor.in_world())	// Initialize main actor.
		init_actors();
					// Paint all the flat scenery.
	for (yoff = starty, cy = start_chunky; cy < stop_chunky; 
						yoff += chunksize, cy++)
		for (xoff = startx, cx = start_chunkx; 
				cx < stop_chunkx; xoff += chunksize, cx++)
			paint_chunk_flats(cx, cy, xoff, yoff);

					// Do each level going upwards.
	for (int lift = 0; lift < 16; lift++)
					// Draw the chunks' objects.
		for (yoff = starty, cy = start_chunky; cy < stop_chunky; 
						yoff += chunksize, cy++)
			for (xoff = startx, cx = start_chunkx; 
				cx < stop_chunkx; xoff += chunksize, cx++)
				paint_chunk_objects(lift, cx, cy, xoff, yoff);
#if 1					/* +++++++Test text drawing.	*/
	static Font_face *font = 0;
	if (!font)
		font = win->open_font("avatar.ttf", 24);
	if (mode == intro)
		{
		win->draw_text_box(font, "Welcome to EXULT V 0.05, a free RPG game engine.\n\nCopyright 1998 J. S. Freedman\nGraphics copyrighted by Origin\nText rendered by FreeType", 15, 15, 600, 400);
		}
#endif
	if (showing_item)		// ID'ing an item?
		win->draw_text(font12, showing_item, showing_rect.x,
							showing_rect.y);
	win->clear_clip();
	painted = 1;
	}

/*
 *	Paint the flat (non-rle) shapes in a chunk.
 */

void Game_window::paint_chunk_flats
	(
	int cx, int cy,			// Chunk coords (0 - 12*16).
	int xoff, int yoff		// Where to place chunk.
	)
	{
	Chunk_object_list& olist = objects[cx][cy];
					// Go through array of shapes.
	for (int shapey = 0; shapey < 16; shapey++)
		for (int shapex = 0; shapex < 16; shapex++)
			{
			ShapeID id = olist.get_flat(shapex, shapey);
			if (!id.is_invalid())
					// Draw shape.
				paint_shape(win, xoff + shapex*8,
					yoff + shapey*8,
					id.get_shapenum(), id.get_framenum());
			}
	}

/*
 *	Paint a chunk's objects, left-to-right, top-to-bottom.
 */

void Game_window::paint_chunk_objects
	(
	int at_lift,			// Only paint this lift.
	int cx, int cy,			// Chunk coords (0 - 12*16).
	int xoff, int yoff		// Where to place chunk.
	)
	{
	Game_object *obj;
	Chunk_object_list& olist = objects[cx][cy];
	for (obj = olist.get_first(); obj; obj = olist.get_next(obj))
		{
		int shapex = obj->get_shape_pos_x();
		int shapey = obj->get_shape_pos_y();
		int lift = obj->get_lift();
		if (lift > at_lift)	// They're sorted by lift first.
			break;
		if (lift != at_lift || lift >= skip_lift)
			continue;	// Don't show things off ground.
					// Is actor inside?  Skip roofs if so.
		if (main_actor_inside && lift >= 5)
			continue;
					// Get shape & frame.
		int shapenum = obj->get_shapenum();
		int framenum = obj->get_framenum();
					// Draw shape.
		paint_shape(win, xoff + shapex*8 - 4*lift, 
				yoff + shapey*8 - 4*lift,
					shapenum, framenum);
		}
	}

/*
 *	Read in a palette.
 */

void Game_window::get_palette
	(
	int pal_num,			// 0-11.
	int brightness			// Brightness % (100 = normal).
	)
	{
	ifstream pal;
	u7open(pal, "static/palettes.flx");
	pal.seekg(256 + 3*256*pal_num);	// Get to desired palette.
	unsigned char colors[3*256];	// Read it in.	
	pal.read(colors, sizeof(colors));
					// They use 6 bits.
	win->set_palette(colors, 63, brightness);
	if (shapewin)
		shapewin->set_palette(colors, 63, brightness);
	}

/*
 *	Brighten/darken palette.
 */

void Game_window::brighten
	(
	int per				// +- percentage to change.
	)
	{
	brightness += per;
	if (brightness < 20)		// Have a min.
		brightness = 20;
	get_palette(0, brightness);
	paint();
	}

/*
 *	Shift view by one chunk.
 */

void Game_window::view_right
	(
	)
	{
	if (chunkx + get_width()/chunksize >= num_chunks - 1)
		return;
					// Shift image to left.
	win->copy(chunksize, 0, get_width() - chunksize, get_height(), 0, 0);
	chunkx++;			// Increment offset.
					// Paint 1 column to right.
	paint(get_width() - chunksize, 0, chunksize, get_height());
	}
void Game_window::view_left
	(
	)
	{
	if (chunkx <= 0)
		return;
	win->copy(0, 0, get_width() - chunksize, get_height(), chunksize, 0);
	chunkx--;
	paint(0, 0, chunksize, get_height());
	}
void Game_window::view_down
	(
	)
	{
	if (chunky + get_height()/chunksize >= num_chunks - 1)
		return;
	win->copy(0, chunksize, get_width(), get_height() - chunksize, 0, 0);
	chunky++;
	paint(0, get_height() - chunksize, get_width(), chunksize);
	}
void Game_window::view_up
	(
	)
	{
	if (chunky < 0)
		return;
	win->copy(0, 0, get_width(), get_height() - chunksize, 0, chunksize);
	chunky--;
	paint(0, 0, get_width(), chunksize);
	}

/*
 *	Animation.
 */

void Game_window::animate
	(
	timeval& time			// Current time of day.
	)
	{
	if (!focus)
		return;			// We're dormant.
	int cx, cy, sx, sy;		// Get chunk, shape within chunk.
	int frame;
	int repaint_all = 0;		// Flag to repaint window.
	if (main_actor.next_frame(time, cx, cy, sx, sy, frame))
		{
					// At left?
		if (cx - chunkx <= 0 && sx < 8)
			view_left();
					// At right?
		else if ((cx - chunkx)*16 + sx >= get_width()/8 - 6)
			view_right();
					// At top?
		if (cy - chunky <= 0 && sy < 8)
			view_up();
					// At bottom?
		else if ((cy - chunky)*16 + sy >= get_height()/8 - 6)
			view_down();
					// Get old chunk it's in.
		int old_cx = main_actor.get_cx(),
					old_cy = main_actor.get_cy();
					// Get old rectangle.
		Rectangle oldrect = get_shape_rect(&main_actor);
					// Move it.
		main_actor.move(cx, cy, &objects[cx][cy], sx, sy, frame);
					// Get new rectangle.
		Rectangle newrect = get_shape_rect(&main_actor);
					// Merge them.
		Rectangle sum = oldrect.add(newrect);
		int inside;		// See if moved inside/outside.
					// In a new chunk?
		if ((main_actor.get_cx() != old_cx ||
		    main_actor.get_cy() != old_cy) &&
			main_actor_inside != find_roof(main_actor.get_cx(),
						main_actor.get_cy()))
			{
			main_actor_inside = !main_actor_inside;
			repaint_all = 1;// Changed, so paint everything.
			}
		else			// Get a little extra.
			paint(sum.x - 4, sum.y - 4, sum.w + 8, sum.h + 8);
		}
	Actor *actor = &main_actor;	// Go through rest of actors.
	while ((actor = actor->get_next()) != &main_actor)
		if (actor->next_frame(time, cx, cy, sx, sy, frame))
			{
					// Get old rectangle.
			Rectangle oldrect = get_shape_rect(actor);
					// Move it.
			actor->move(cx, cy, &objects[cx][cy], sx, sy, frame);
			if (repaint_all)// Already going to do a full repaint?
				continue;
					// Get new rectangle.
			Rectangle newrect = get_shape_rect(actor);
					// Merge them.
			Rectangle sum = oldrect.add(newrect);
			sum.x -= 4;	// Make a little bigger.
			sum.y -= 4;
			sum.w += 8;
			sum.h += 8;
					// Intersect with screen.
			Rectangle winrect = get_win_rect();
			sum = sum.intersect(winrect);
					// Watch for negatives.
			if (sum.w > 0 && sum.h > 0)
				paint(sum.x, sum.y, sum.w, sum.h);
			}
	if (repaint_all)
		paint();		// Paint whole window now.
	}

/*
 *	Start moving the actor.
 */

void Game_window::start_actor
	(
	int winx, int winy		// Move towards this win. location.
	)
	{
	if (mode != normal)
		return;
					// Move every 1/8 sec.
	main_actor.start(chunkx*chunksize + winx, chunky*chunksize + winy,
						125000);
	}

/*
 *	Stop the actor.
 */

void Game_window::stop_actor
	(
	)
	{
	main_actor.stop();		// Stop and set resting state.
	paint();
	}

/*
 *	Find a "roof" in the given chunk.
 *
 *	Output:	1 if found, else 0.
 */

int Game_window::find_roof
	(
	int cx, int cy			// Absolute chunk coords.
	)
	{
	Game_object *obj;
	Chunk_object_list& olist = objects[cx][cy];
	for (obj = olist.get_first(); obj; obj = olist.get_next(obj))
		if (obj->get_lift() >= 5)
			return (1);	// Found one.
	return (0);
	}

/*
 *	Find objects at a given position on the screen with a given lift.
 *
 *	Output:	# of objects, stored in list.
 */

int Game_window::find_objects
	(
	int lift,			// Look for objs. with this lift.
	int x, int y,			// Pos. on screen.
	Game_object **list		// Objects found are stored here.
	)
	{
//	cout << "Find_objects:  mouse is at (" << x << ',' << y << ")\n";
					// Figure chunk #'s.
	int start_cx = chunkx + (x + 4*lift)/chunksize;
	int start_cy = chunky + (y + 4*lift)/chunksize;
	Game_object *obj;
	int cnt = 0;			// Count # found.
					// Check 1 chunk down & right too.
	for (int ycnt = 0; ycnt < 2; ycnt++)
		{
		int cy = start_cy + ycnt;
		for (int xcnt = 0; xcnt < 2; xcnt++)
			{
			int cx = start_cx + xcnt;
			Chunk_object_list& olist = objects[cx][cy];
			for (obj = olist.get_first(); obj; 
						obj = olist.get_next(obj))
				{
				if (obj->get_lift() != lift)
					continue;
				Rectangle r = get_shape_rect(obj, cx, cy);
				if (r.has_point(x, y))
					list[cnt++] = obj;
				}
			}
		}
	return (cnt);
	}

/*
 *	Show the names of the items the mouse is clicked on.
 */

void Game_window::show_items
	(
	int x, int y			// Coords. in window.
	)
	{
	Game_object *found[100];
	int cnt = 0;
					// See what was clicked on.
	for (int lift = 0; lift < 3; lift++)
		cnt += find_objects(main_actor.get_lift() + lift, 
							x, y, &found[cnt]);
	for (int i = 0; i < cnt; i++)	// Go through them.
		{
		Game_object *obj = found[i];
					// Show name.
		char *item_name = item_names[obj->get_shapenum()];
		if (item_name)
			{		// Just do the first one found.
			showing_item = item_name;
			showing_rect.x = x;
			showing_rect.y = y;
			showing_rect.w = win->get_text_width(
							font12, item_name);
			showing_rect.h = win->get_text_height(font12);
			paint(showing_rect.x, showing_rect.y,
				showing_rect.w + 8, showing_rect.h + 8);
			return;
			}
		}
	}

/*
 *	Handle a double-click.
 */

void Game_window::double_clicked
	(
	Window xwin,			// Window it occurred in.
	int x, int y			// Coords in window.
	)
	{
	if (xwin != win->get_win())
		return;			// Not the main window.
	Game_object *found[100];
	int cnt = 0;
					// See what was clicked on.
	for (int lift = 0; lift < 3; lift++)
		cnt += find_objects(main_actor.get_lift() + lift, 
							x, y, &found[cnt]);
//	cout << cnt << " objects found.\n";
	for (int i = 0; i < cnt; i++)	// Go through them.
		{
		Game_object *obj = found[i];
		switch (obj->get_shapenum())
			{
		case CLOSED_DOOR:
			obj->set_shape(OPEN_DOOR); paint();
			break;
		case OPEN_DOOR:
			obj->set_shape(CLOSED_DOOR); paint();
			break;
		case CLOSED_HORIZ_SHUTTERS:
			obj->set_shape(OPEN_HORIZ_SHUTTERS); paint();
			break;
		case OPEN_HORIZ_SHUTTERS:
			obj->set_shape(CLOSED_HORIZ_SHUTTERS); paint();
			break;
		case CLOSED_VERT_SHUTTERS:
			obj->set_shape(OPEN_VERT_SHUTTERS); paint();
			break;
		case OPEN_VERT_SHUTTERS:
			obj->set_shape(CLOSED_VERT_SHUTTERS); paint();
			break;
			}
		}
	}

/*
 *	Globals:
 */
Game_window *gwin = 0;
Atom wm_delete_window, wm_protocols;

static void Handle_event(Display *, Game_window *);

/*
 *	Main program.
 */

int main
	(
	int argc,
	char *argv[]
	)
	{
	cout << "Exult V0.05.  Copyright (C) 1998 J. S. Freedman\n";
	cout << "Text rendering done using the 'FreeType' font engine.\n";
					// Open display.
	Display *display = XOpenDisplay(NULL);
	if (!display)
		{
		cerr << "Can't open display.\n";
		exit(-1);
		}
	gwin = new Game_window(display);
					// Want to know if win. is closed.
	wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(display, gwin->get_win()->get_win(),
							&wm_delete_window, 1);
	wm_protocols = XInternAtom(display, "WM_PROTOCOLS", False);

	if (argc > 2)			// Specify chunkx, chunky on cmndline.
		gwin->set_chunk_offsets(atoi(argv[1]), atoi(argv[2]));
	else				// Else start in Trinsic.
		gwin->set_chunk_offsets(64, 135);
					// Get connection number.
	int xfd = ConnectionNumber(display);
	/*
	 *	Main event loop.
	 */
	while (!gwin->is_quitting_time())
		{
		fd_set rfds;		// Want a timer.
		struct timeval timer;
		timer.tv_sec = 0;
		timer.tv_usec = 20000;	// Try 1/50 second.
		FD_ZERO(&rfds);
		FD_SET(xfd, &rfds);
					// Wait for timeout or event.
		select(xfd + 1, &rfds, 0, 0, &timer);
		while (!gwin->is_quitting_time() && XPending(display) > 0)
			Handle_event(display, gwin);
					// Get current time.
		gettimeofday(&timer, 0);
		gwin->animate(timer);
		gwin->show();		// Blit to screen if necessary.
		}
	delete gwin;
	XCloseDisplay(display);		// All done.
	return (0);
	}

/*
 *	Handle an event.
 */

static void Handle_event
	(
	Display *display,
	Game_window *gwin
	)
	{
	static int shape_cnt = 254, shape_frame = 0;
	static Time last_b1_click = 0;	// For detecting double-clicks.
	XEvent event;
	Window root, win;		// Get window ID's.
	int rootx, rooty, winx, winy;	// Get positions.
	unsigned int mask;
	char keybuf[10];
	KeySym keysym;
	XComposeStatus compose;
	XNextEvent(display, &event);
	switch (event.type)
		{
	case ButtonPress:
		gwin->end_intro();
		gwin->stop_showing_item();
					// Move sprite toward mouse
					//  when right button pressed.
		if (event.xbutton.button == 3)
			gwin->start_actor(event.xbutton.x, event.xbutton.y);
#if 1
		else if (event.xbutton.button == 2)
			gwin->debug_shape(event.xbutton.window,
					event.xbutton.x, event.xbutton.y);
#endif
		break;
	case ButtonRelease:
		if (event.xbutton.button == 3)
			gwin->stop_actor();
		else if (event.xbutton.button == 1)
			{		// Last click within .5 secs?
			if (event.xbutton.time - last_b1_click < 500)
				{
				gwin->double_clicked(event.xbutton.window,
					event.xbutton.x, event.xbutton.y);
				break;
				}
			last_b1_click = event.xbutton.time;
					// Identify item(s) clicked on.
			gwin->show_items(event.xbutton.x, event.xbutton.y);
			}
		break;
	case MotionNotify:		// Moving with right button down.
		while (XCheckMaskEvent(display, Button3MotionMask, &event))
			;
		if (XQueryPointer(display, event.xmotion.window, &root,
				&win, &rootx, &rooty, &winx, &winy, &mask))
			gwin->start_actor(winx, winy);
		break;
	case Expose:			// Just repaint on last expose.
		if (event.xexpose.count == 0)
			gwin->show(event.xexpose.window);
		break;
	case FocusIn:
		gwin->get_focus(event.xexpose.window);
		break;
	case FocusOut:
		gwin->lose_focus(event.xexpose.window);
		break;
	case ConfigureNotify:		// Resize.
		gwin->resized(event.xconfigure.window,
			event.xconfigure.width, event.xconfigure.height);
		break;
	case ClientMessage:
		cout << "ClientMessage\n";
		if (event.xclient.format == 32 &&
			event.xclient.message_type == wm_protocols &&
			event.xclient.data.l[0] == (long) wm_delete_window)
				gwin->set_quitting_time();
		break;
	case DestroyNotify:
		cout << "DestroyNotify\n";
		break;
	case MapNotify:
		cout << "MapNotify\n";
		break;
	case UnmapNotify:
		cout << "UnmapNotify\n";
		break;
	case KeyPress:			// Keystroke.
		XLookupString((XKeyEvent *) &event, 
			keybuf, sizeof(keybuf), &keysym, &compose);
		switch (keysym)
			{
		case XK_plus:		// Brighten.
			gwin->brighten(20);
			break;
		case XK_minus:		// Darken.
			gwin->brighten(-20);
			break;
		case XK_b:
			Breakpoint();
			break;
		case XK_q:
			gwin->set_quitting_time();
			break;
		case XK_l:		// Decrement skip_lift.
			if (gwin->skip_lift == 16)
				gwin->skip_lift = 8;
			else
				gwin->skip_lift--;
			if (gwin->skip_lift == 0)
				gwin->skip_lift = 16;
			cout << "Skip_lift = " << gwin->skip_lift
						<< '\n';
					// FALL THROUGH.
		case XK_p:		// Rerender image.
			gwin->paint();
			break;
#if 1
		case XK_s:		// Show next shape.
			shape_frame = 0;
			if (++shape_cnt == gwin->get_num_shapes())
				shape_cnt = 0;
			cout << "Painting shape " << shape_cnt << '\n';
			gwin->paint();
			gwin->paint_shape(gwin->get_win(),
				200, 200, shape_cnt, shape_frame);
			break;
		case XK_S:		// Show prev. shape.
			shape_frame = 0;
			if (--shape_cnt < 0)
				shape_cnt = gwin->get_num_shapes() - 1;
			cout << "Painting shape " << shape_cnt << '\n';
			gwin->paint();
			gwin->paint_shape(gwin->get_win(),
				200, 200, shape_cnt, shape_frame);
			break;
		case XK_f:		// Show next frame.
			cout << "Frame # " << ++shape_frame << '\n';
			gwin->paint();
			gwin->paint_shape(gwin->get_win(),
				200, 200, shape_cnt, shape_frame);
			break;
#endif
		case XK_Right:
			gwin->view_right();
			break;
		case XK_Left:
			gwin->view_left();
			break;
		case XK_Down:
			gwin->view_down();
			break;
		case XK_Up:
			gwin->view_up();
			break;
			}
		break;
		}
	}
