/**
 **	Exult.h - X-windows Ultima7 map browser.
 **
 **	Written: 7/22/98 - JSF
 **/

/*
Copyright (C) 1998  Jeffrey S. Freedman

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

class Font_face;

/*
 *	The main game window:
 */
class Game_window
	{
	Display *display;		// X-windows display.
	int screen_num;
	Image_window *win;		// Window to display into.
	Image_window *shapewin;		// Debug win. for showing shapes.
	Game_object shapewin_objs[20];	// List of shapes to show.
	int shapewin_cur;		// Index of current shape being shown
					//   in shapewin.
	Font_face *font12;		// 12-point "avatar.ttf" font.
public:
	enum Game_mode {		// Can be in different modes.
		intro,			// Splash screen.
		normal,			// Normal game-play.
		conversation		// Talking.
		};
private:
	Game_mode mode;			// Mode we'er in.
	char *showing_item;		// Item we're showing the name of.
	Rectangle showing_rect;		// Rectangle text is shown in.
	unsigned char quitting_time;	// Time to quit.
	unsigned char painted;		// 1 if we updated image buffer.
	unsigned char focus;		// Do we have focus?
	ifstream chunks;		// "u7chunks" file.
	ifstream shapes;		// "shapes.vga" file.
	ifstream u7map;			// "u7map" file.
	int num_shapes;			// Total # of shapes.
	Shape ***shapes_read;		// List of ->'s to shapes' lists
					//   of frames in memory.
	Human_actor main_actor;		// Main sprite to move around.
	unsigned char main_actor_inside;// 1 if actor is in a building.
					// A list of objects in each chunk.
	Chunk_object_list objects[num_chunks][num_chunks];
	unsigned char schunk_read[144];	// Flag for reading in each "ifix".
	int chunkx, chunky;		// Chunk coord. of window within world.
	int brightness;			// Palette brightness.
					// Open a U7 file.
	int u7open(ifstream& in, char *fname, int dont_abort = 0);
public:
	int skip_lift;			// Skip objects with lift > 0.
	int debug;
	Game_window(Display *dpy);
	~Game_window();
	void open_shape_window();	// Open window to show shapes.
	void abort(char *msg, ...);	// Fatal error.
	int get_width()
		{ return win->get_width(); }
	int get_height()
		{ return win->get_height(); }
	Rectangle get_win_rect()	// Get window's rectangle.
		{ return Rectangle(0, 0, win->get_width(), win->get_height());}
	Image_window *get_win()
		{ return win; }
	int get_num_shapes()
		{ return num_shapes; }
	int is_quitting_time()
		{ return quitting_time; }
	void set_quitting_time()
		{ quitting_time = 1; }
	void set_mode(Game_mode md)
		{ mode = md; }
					// Resize event occurred.
	void resized(Window xwin, unsigned int neww, unsigned int newh);
	void show()
		{ 
		if (painted)
			{
			win->show();
			painted = 0;
			}
		}
	void show(Window xwin)
		{
		if (shapewin && xwin == shapewin->get_win())
			shapewin->show();
		else
			win->show();
		}
	void set_chunk_offsets(int cx, int cy)
		{
		if (cx >= 0 && cx < num_chunks)
			chunkx = cx;
		if (cy >= 0 && cy < num_chunks)
			chunky = cy;
		}
#if 1
					// Show shape(s) clicked on.
	void debug_shape(Window xwin, int x, int y);	// DEBUGGING.
	void find_debug_shapes(int lift, int x, int y);
	void shapewin_paint(int num);
#endif
					// Get shape.
	Shape *get_shape(int shapenum, int framenum)
		{
					// Already in memory?
		Shape **frames = shapes_read[shapenum];
		Shape *shape = frames ? frames[framenum] : 0;
		if (!shape)		// If not, read it in.
			{
			shape = new Shape();
			int framecnt = shape->read(shapes, shapenum, framenum);
			if (framenum >= framecnt)
				{
				delete shape;
				return (0);
				}
			if (!frames)	// First one?
				{
				shapes_read[shapenum] = 
					frames = new Shape *[framecnt];
				memset((char *) frames, 0, 
						framecnt * sizeof(Shape *));
				}
			frames[framenum] = shape;
			}
		return (shape);
		}
	Shape *get_shape(ShapeID& id)
		{ return get_shape(id.get_shapenum(), id.get_framenum()); }
					// Get screen area used by object in
					//   given chunk.
	Rectangle get_shape_rect(Game_object *obj, int cx, int cy)
		{
		Shape *s = get_shape(*obj);
		int lft = 4*obj->get_lift();
		return Rectangle(
			(cx - chunkx)*chunksize +
				obj->get_shape_pos_x()*8 + 8 - s->xleft - lft,
			(cy - chunky)*chunksize +
				obj->get_shape_pos_y()*8 + 8 - s->yabove - lft,
			s->xleft + s->xright,
			s->yabove + s->ybelow
			);
		}

					// Get screen area used by sprite.
	Rectangle get_shape_rect(Sprite *sprite)
		{ return get_shape_rect(sprite, sprite->get_cx(),
							sprite->get_cy()); }
					// Paint shape in window.
	void paint_shape(Image_window *iwin,
			int xoff, int yoff, int shapenum, int framenum);
					// Read encoded show into window.
	void get_rle_shape(Image_window *iwin,
					Shape& shape, int xoff, int yoff);
					// Get "map" superchunk objs/scenery.
	void get_map_objects(int schunk);
					// Get "chunk" objects/scenery.
	void get_chunk_objects(int cx, int cy, int chunk_num);
					// Get "ifix" objects for a superchunk.
	void get_ifix_objects(int schunk);
					// Get "ifix" objs. for given chunk.
	void get_ifix_chunk_objects(ifstream& ifix, long filepos, int cnt,
							int cx, int cy);
					// Get moveable objects.
	void get_ireg_objects(int schunk);
					// Get all superchunk objects.
	void get_superchunk_objects(int schunk);
	void init_actors();		// Place actors in the world.
					// Paint area of image.
	void paint(int x, int y, int w, int h);
	void paint()			// Paint whole image.
		{ paint(0, 0, get_width(), get_height()); }
					// Paint "flat" scenery in a chunk.
	void paint_chunk_flats(int cx, int cy, int xoff, int yoff);
					// Paint objects in given chunk at
					//   given lift.
	void paint_chunk_objects(int at_lift,
				int cx, int cy, int xoff, int yoff);
					// Get desired palette.
	void get_palette(int pal_num, int brightness = 100);
	void brighten(int per);		// Brighten/darken by percentage.
	void view_right();		// Move view 1 chunk to right.
	void view_left();		// Move view left by 1 chunk.
	void view_down();		// Move view down.
	void view_up();			// Move view up.
	void animate(timeval& time);	// Do animation at given time.
					// Start moving actor.
	void start_actor(int winx, int winy);
	void stop_actor();		// Stop moving the actor.
	int find_roof(int cx, int cy);	// Find a "roof" in given chunk.
					// Find objects that (x,y) is in.
	int find_objects(int lift, int x, int y, Game_object **list);
	void show_items(int x, int y);	// Show names of items clicked on.
					// Handle a double-click in window.
	void double_clicked(Window xwin, int x, int y);
	void get_focus(Window xwin)	// Get/lose focus.
		{
		if (xwin == win->get_win())
			focus = 1;
		}
	void lose_focus(Window xwin)
		{
		if (xwin == win->get_win())
			focus = 0;
		}
	void end_intro()		// End splash screen.
		{
		if (mode == intro)
			{
			mode = normal;
			paint();
			}
		}
	void stop_showing_item()	// Turn off display of item name.
		{
		if (showing_item)
			{
			showing_item = 0;
			paint(showing_rect.x - 4, showing_rect.y - 4,
				showing_rect.w + 8, showing_rect.h + 8);
			}
		}
	};

